<?php

namespace Drupal\sendethic\Form;

use Drupal\Component\Render\FormattableMarkup;
use Drupal\Core\Form\ConfigFormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Render\Markup;
use Drupal\sendethic\SendEthicAccount;
use Drupal\sendethic\SendEthicAttributeOption;
use Drupal\sendethic\SendEthicSegment;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Configure SendEthic settings for this site.
 */
class SettingsForm extends ConfigFormBase {

  /**
   * The sendethic manager.
   *
   * @var \Drupal\sendethic\Service\SendEthicManager
   */
  protected $sendEthicManager;

  /**
   * The module handler.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected $moduleHandler;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    $instance = parent::create($container);
    $instance->sendEthicManager = $container->get('sendethic.manager');
    $instance->moduleHandler = $container->get('module_handler');
    return $instance;
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'sendethic_settings';
  }

  /**
   * {@inheritdoc}
   */
  protected function getEditableConfigNames() {
    return ['sendethic.settings'];
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    // Credentials
    $originalConfig = \Drupal::configFactory()->get('sendethic.settings');
    $accountId = $this->config('sendethic.settings')->get('default.account_id');
    $realAccountId = $originalConfig->hasOverrides('default.account_id') ? $originalConfig->get('default.account_id') : $accountId;
    $form['account_id'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Account ID'),
      '#default_value' => $realAccountId,
      '#disabled' => $originalConfig->hasOverrides('default.account_id'),
      '#description' => $originalConfig->hasOverrides('default.account_id') ? 'Overidden (by Key module)' : '',
    ];
    $apiKey = $this->config('sendethic.settings')->get('default.api_key');
    $realApiKey = $originalConfig->hasOverrides('default.api_key') ? $originalConfig->get('default.api_key') : $apiKey;
    $form['api_key'] = [
      '#type' => 'textfield',
      '#title' => $this->t('API key'),
      '#default_value' => $realApiKey,
      '#disabled' => $originalConfig->hasOverrides('default.api_key'),
      '#description' => $originalConfig->hasOverrides('default.api_key') ? 'Overidden (by Key module)' : '',
    ];
    // dump($this->config('sendethic.settings')->getRawData(), $accountId, $apiKey);

    $account = $realAccountId && $realApiKey ? new SendEthicAccount($realAccountId, $realApiKey) : null;
    $segments = $segmentOptions = $attributes = $attributesOptions = [];
    if ($account instanceof SendEthicAccount) {
      $form['account_status'] = [
        '#markup' => new FormattableMarkup('<strong style="color:green">@account</strong>', [
          '@account' => $this->t('Account @account connected', ['@account' => $account->getAccountId()]),
        ]),
      ];
      $segments = $this->sendEthicManager->getSegments($account);
      foreach ($segments as $segment) {
        $segmentOptions[$segment->getId()] = sprintf('%s (%s)', $segment->getName(), $segment->getDetails());
      }
      $attributes = $this->sendEthicManager->getAttributes($account);
      foreach ($attributes as $attribute) {
        $attributesOptions[$attribute->getId()] = sprintf('%s (%s - %s)', $attribute->getName(), $attribute->getFormat(), $attribute->getDescription() ?: '');
      }

      // Attributes
      $form['attributes'] = [
        '#markup' => new FormattableMarkup('<h6>@label</h6>', [
          '@label' => $this->t('Attributes (default values)'),
        ]),
        '#tree' => true,
      ];
      foreach ($attributes as $attribute) {
        $attributeValue = $this->config('sendethic.settings')->get(sprintf('default.attributes.%s', $attribute->getId()));
        if (in_array($attribute->getFormat(), ['SingleSelection', 'MultipleSelection'])) {
          $options = [];
          $attributeValue = is_string($attributeValue) ? array_filter(explode(',', $attributeValue)) : $attributeValue;
          foreach ($attribute->getOptions() as $option) {
            /** @var SendEthicAttributeOption $option */
            $options[$option->getId()] = sprintf('%s (ID: %d)', $option->getName(), $option->getId());
          }
          $form['attributes'][$attribute->getId()] = [
            '#type' => $this->moduleHandler->moduleExists('select2') ? 'select2' : 'select',
            '#title' => sprintf('%s (ID: %d - Format: %s)', $attribute->getName(), $attribute->getId(), $attribute->getFormat()),
            '#options' => $options,
            '#empty_option' => $this->t('- Select -'),
            '#description' => $attribute->getDescription(),
            '#default_value' => $attributeValue,
            '#multiple' => $attribute->getFormat() === 'MultipleSelection',
          ];
        } else {
          $form['attributes'][$attribute->getId()] = [
            '#type' => $attribute->getFormat() === 'Number' ? 'number' : 'textfield',
            '#title' => sprintf('%s (ID: %d - Format: %s)', $attribute->getName(), $attribute->getId(), $attribute->getFormat()),
            '#description' => $attribute->getDescription(),
            '#default_value' => $attributeValue,
          ];
        }
      }

      // Segments
      $form['segment'] = [
        '#markup' => new FormattableMarkup('<h6>@label</h6><ul>@list</ul>', [
          '@label' => $this->t('Segments available'),
          '@list' => Markup::create(join('', array_map(function(SendEthicSegment $segment) {
            return (new FormattableMarkup('<li><strong>@label</strong><br /><small>@details</small></li>', [
              '@label' => $segment->getName(),
              '@details' => $segment->getDetails(),
            ]))->__toString();
          }, $segments))),
        ]),
      ];
    }

    return parent::buildForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state) {
    // if ($form_state->getValue('account_id') != 'example') {
    //   $form_state->setErrorByName('account_id', $this->t('The value is not correct.'));
    // }
    // if ($form_state->getValue('api_key') != 'example') {
    //   $form_state->setErrorByName('api_key', $this->t('The value is not correct.'));
    // }
    parent::validateForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $attributes = [];
    foreach (($form_state->getValue('attributes') ?: []) as $attributeId => $attributeValue) {
      if (!empty($attributeValue)) {
        $attributes[$attributeId] = is_array($attributeValue) ? join(',', $attributeValue) : $attributeValue;
      }
    }
    $this->config('sendethic.settings')
      ->set('default.account_id', $form_state->getValue('account_id'))
      ->set('default.api_key', $form_state->getValue('api_key'))
      ->set('default.attributes', $attributes)
      ->save();
    parent::submitForm($form, $form_state);
  }

}
