<?php

namespace Drupal\sendethic\Plugin\WebformHandler;

use Drupal\Component\Render\FormattableMarkup;
use Drupal\Component\Utility\Html;
use Drupal\Core\Ajax\AjaxResponse;
use Drupal\Core\Ajax\RedirectCommand;
use Drupal\Core\EventSubscriber\MainContentViewSubscriber;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\Routing\TrustedRedirectResponse;
use Drupal\Core\Serialization\Yaml;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\sendethic\SendEthicAccount;
use Drupal\sendethic\SendEthicContact;
use Drupal\sendethic\Service\SendEthicManager;
use Drupal\webform\Element\WebformAjaxElementTrait;
use Drupal\webform\Element\WebformCompositeBase;
use Drupal\webform\Element\WebformSelectOther;
use Drupal\webform\Plugin\WebformHandlerBase;
use Drupal\webform\Utility\WebformOptionsHelper;
use Drupal\webform\WebformMessageManagerInterface;
use Drupal\webform\WebformSubmissionInterface;
use GuzzleHttp\Exception\RequestException;
use Psr\Http\Message\ResponseInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * SendEthic Webform Handler.
 *
 * @WebformHandler(
 *   id = "sendethic",
 *   label = @Translation("SendEthic"),
 *   category = @Translation("External"),
 *   description = @Translation("SendEthic custom webform submission handler."),
 *   cardinality = \Drupal\webform\Plugin\WebformHandlerInterface::CARDINALITY_UNLIMITED,
 *   results = \Drupal\webform\Plugin\WebformHandlerInterface::RESULTS_IGNORED,
 *   submission = \Drupal\webform\Plugin\WebformHandlerInterface::SUBMISSION_REQUIRED,
 * )
 */
class SendEthicWebformHandler extends WebformHandlerBase {

  use WebformAjaxElementTrait;
  use StringTranslationTrait;

  /**
   * Default value. (This is used by the handler's settings.)
   */
  const DEFAULT_EMAIL_VALUE = '_default';

  /**
   * The sendethic manager.
   *
   * @var \Drupal\sendethic\Service\SendEthicManager
   */
  protected $sendEthicManager;

  /**
   * The token manager.
   *
   * @var \Drupal\webform\WebformTokenManagerInterface
   */
  protected $tokenManager;

  /**
   * The module handler.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected $moduleHandler;

  /**
   * The webform message manager.
   *
   * @var \Drupal\webform\WebformMessageManagerInterface
   */
  protected $messageManager;

  /**
   * The webform element plugin manager.
   *
   * @var \Drupal\webform\Plugin\WebformElementManagerInterface
   */
  protected $elementManager;

  /**
   * List of unsupported webform submission properties.
   *
   * The below properties will not being included in a remote post.
   *
   * @var array
   */
  protected $unsupportedProperties = [
    'metatag',
  ];

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    $instance = parent::create($container, $configuration, $plugin_id, $plugin_definition);
    $instance->sendEthicManager = $container->get('sendethic.manager');
    $instance->tokenManager = $container->get('webform.token_manager');
    $instance->moduleHandler = $container->get('module_handler');
    $instance->messageManager = $container->get('webform.message_manager');
    $instance->elementManager = $container->get('plugin.manager.webform.element');
    return $instance;
  }

  /**
   * {@inheritdoc}
   */
  public function getSummary() {
    $settings = $this->getSettings();
    $settings['contact_mail'] = ($this->configuration['contact_mail'] === static::DEFAULT_EMAIL_VALUE) ? '?' : $this->configuration['contact_mail'];
    $settings['contact_firstname'] = $this->configuration['contact_firstname'] ?: '?';
    $settings['contact_lastname'] = $this->configuration['contact_lastname'] ?: '?';
    return [
      '#settings' => $settings,
    ] + parent::getSummary();
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    // We can't inject the entity field manager dependency because
    // RemotePostWebformHandler::defaultConfiguration() is called with in
    // RemotePostWebformHandler::create().
    // @see \Drupal\webform\Plugin\WebformHandlerBase::create
    // @see https://www.drupal.org/project/webform/issues/3285846
    /** @var \Drupal\Core\Entity\EntityFieldManagerInterface $entity_field_manager */
    $entity_field_manager = \Drupal::service('entity_field.manager');
    $field_names = array_keys($entity_field_manager->getBaseFieldDefinitions('webform_submission'));
    $excluded_data = array_combine($field_names, $field_names);
    return [
      'account' => 'default',
      'contact_mail' => static::DEFAULT_EMAIL_VALUE,
      'contact_firstname' => '',
      'contact_lastname' => '',
      'attributes' => [],
      'excluded_data' => $excluded_data,
      'custom_data' => '',
      'debug' => FALSE,
      // Custom response messages.
      'message' => '',
      'messages' => [],
      // Custom response redirect URL.
      'error_url' => '',
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $webform = $this->getWebform();

    // Options
    $name_element_options = [];
    $mail_element_options = [];

    $elements = $this->webform->getElementsInitializedAndFlattened();
    foreach ($elements as $element_key => $element) {
      $element_plugin = $this->elementManager->getElementInstance($element);
      if (!$element_plugin->isInput($element) || !isset($element['#type'])) {
        continue;
      }

      // Set title.
      $element_title = (isset($element['#title'])) ? new FormattableMarkup('@title (@key)', ['@title' => $element['#title'], '@key' => $element_key]) : $element_key;

      // Multiple value elements can NOT be used as a tokens.
      if ($element_plugin->hasMultipleValues($element)) {
        continue;
      }

      if (!$element_plugin->isComposite()) {
        // Add name element token.
        $name_element_options["[webform_submission:values:$element_key:raw]"] = $element_title;

        // Add mail element token.
        if (in_array($element['#type'], ['email', 'hidden', 'value', 'textfield', 'webform_email_multiple', 'webform_email_confirm'])) {
          $mail_element_options["[webform_submission:values:$element_key:raw]"] = $element_title;
        }
      }

      // Element type specific tokens.
      switch ($element['#type']) {
        case 'webform_name':
          // Allow 'webform_name' composite to be used a value token.
          $name_element_options["[webform_submission:values:$element_key:value]"] = $element_title;
          break;
      }

      // Handle composite sub elements.
      if ($element_plugin instanceof WebformCompositeBase) {
        $composite_elements = $element_plugin->getCompositeElements();
        foreach ($composite_elements as $composite_key => $composite_element) {
          $composite_element_plugin = $this->elementManager->getElementInstance($element);
          if (!$composite_element_plugin->isInput($element) || !isset($composite_element['#type'])) {
            continue;
          }

          // Set composite title.
          if (isset($element['#title'])) {
            $f_args = [
              '@title' => $element['#title'],
              '@composite_title' => $composite_element['#title'],
              '@key' => $element_key,
              '@composite_key' => $composite_key,
            ];
            $composite_title = new FormattableMarkup('@title: @composite_title (@key: @composite_key)', $f_args);
          }
          else {
            $composite_title = "$element_key:$composite_key";
          }

          // Add name element token. Only applies to basic (not composite) elements.
          $name_element_options["[webform_submission:values:$element_key:$composite_key:raw]"] = $composite_title;

          // Add mail element token.
          if (in_array($composite_element['#type'], ['email', 'webform_email_multiple', 'webform_email_confirm'])) {
            $mail_element_options["[webform_submission:values:$element_key:$composite_key:raw]"] = $composite_title;
          }
        }
      }
    }

    // Get email and name other.
    $other_element_email_options = [
      '[site:mail]' => 'Site email address',
      '[current-user:mail]' => 'Current user email address [Authenticated only]',
      '[webform:author:mail]' => 'Webform author email address',
      '[webform_submission:user:mail]' => 'Webform submission owner email address [Authenticated only]',
    ];
    $other_element_name_options = [
      '[site:name]' => 'Site name',
      '[current-user:display-name]' => 'Current user display name',
      '[current-user:account-name]' => 'Current user account name',
      '[webform:author:display-name]' => 'Webform author display name',
      '[webform:author:account-name]' => 'Webform author account name',
      '[webform_submission:user:display-name]' => 'Webform submission author display name',
      '[webform_submission:user:account-name]' => 'Webform submission author account name',
    ];

    // Additional.
    $form['additional'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('Additional settings'),
    ];
    $form['additional']['custom_data'] = [
      '#type' => 'webform_codemirror',
      '#mode' => 'yaml',
      '#title' => $this->t('Custom data'),
      '#description' => $this->t('Enter custom data that will be included in all remote post requests.'),
      '#default_value' => $this->configuration['custom_data'],
    ];
    $form['additional']['message'] = [
      '#type' => 'webform_html_editor',
      '#title' => $this->t('Custom error response message'),
      '#description' => $this->t('This message is displayed when the response status code is not 2xx.') . '<br/><br/>' . $this->t('Defaults to: %value', ['%value' => $this->messageManager->render(WebformMessageManagerInterface::SUBMISSION_EXCEPTION_MESSAGE)]),
      '#default_value' => $this->configuration['message'],
    ];
    $form['additional']['messages_token'] = [
      '#type' => 'webform_message',
      '#message_message' => $this->t('Response data can be passed to response message using [webform:handler:{machine_name}:{key}] tokens. (i.e. [webform:handler:remote_post:message])'),
      '#message_type' => 'info',
    ];
    $form['additional']['messages'] = [
      '#type' => 'webform_multiple',
      '#title' => $this->t('Custom response messages'),
      '#description' => $this->t('Enter custom response messages for specific status codes.'),
      '#empty_items' => 0,
      '#no_items_message' => $this->t('No error response messages entered. Please add messages below.'),
      '#add' => FALSE,
      '#element' => [
        'code' => [
          '#type' => 'webform_select_other',
          '#title' => $this->t('Response status code'),
          '#options' => [
            '200' => $this->t('200 OK'),
            '201' => $this->t('201 Created'),
            '204' => $this->t('204 No Content'),
            '400' => $this->t('400 Bad Request'),
            '401' => $this->t('401 Unauthorized'),
            '403' => $this->t('403 Forbidden'),
            '404' => $this->t('404 Not Found'),
            '500' => $this->t('500 Internal Server Error'),
            '502' => $this->t('502 Bad Gateway'),
            '503' => $this->t('503 Service Unavailable'),
            '504' => $this->t('504 Gateway Timeout'),
          ],
          '#other__type' => 'number',
          '#other__description' => $this->t('<a href="https://en.wikipedia.org/wiki/List_of_HTTP_status_codes">List of HTTP status codes</a>.'),
        ],
        'message' => [
          '#type' => 'webform_html_editor',
          '#title' => $this->t('Response message'),
        ],
      ],
      '#default_value' => $this->configuration['messages'],
    ];
    $form['additional']['error_url'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Custom error response redirect URL'),
      '#description' => $this->t('The URL or path to redirect to when a remote fails.'),
      '#default_value' => $this->configuration['error_url'],
      '#pattern' => '(https?:\/\/|\/).+',
    ];

    // Development.
    $form['development'] = [
      '#type' => 'details',
      '#title' => $this->t('Development settings'),
    ];
    $form['development']['debug'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Enable debugging'),
      '#description' => $this->t('If checked, posted submissions will be displayed onscreen to all users.'),
      '#return_value' => TRUE,
      '#default_value' => $this->configuration['debug'],
    ];

    // SendEthic contact information data.
    $form['sendethic_data'] = [
      '#type' => 'details',
      '#title' => $this->t('SendEthic contact'),
      '#open' => true,
    ];

    $form['sendethic_data']['contact_mail'] = $this->buildElement('contact_mail', $this->t('Contact email'), $this->t('Contact email address'), TRUE, $mail_element_options, $other_element_email_options);
    $form['sendethic_data']['contact_firstname'] = $this->buildElement('contact_firstname', $this->t('Contact first name'), $this->t('Contact first name'), FALSE, $name_element_options, $other_element_name_options);
    $form['sendethic_data']['contact_lastname'] = $this->buildElement('contact_lastname', $this->t('Contact last name'), $this->t('Contact last name'), FALSE, $name_element_options, $other_element_name_options);

    $form['sendethic_data']['account'] = [
      '#type' => 'select',
      '#title' => $this->t('Account'),
      '#options' => [
        'default' => $this->t('Default'),
      ],
      '#default_value' => $this->configuration['account'],
      '#required' => true,
    ];

    $account = $this->sendEthicManager->getDefaultAccount();
    if ($account instanceof SendEthicAccount) {
      // Attributes
      $attributes = $this->sendEthicManager->getAttributes($account);
      foreach ($attributes as $attribute) {
        $attributesOptions[$attribute->getId()] = sprintf('%s (%s - %s)', $attribute->getName(), $attribute->getFormat(), $attribute->getDescription() ?: '');
      }
      $form['sendethic_data']['attributes'] = [
        '#markup' => new FormattableMarkup('<h6>@label</h6>', [
          '@label' => $this->t('Attributes'),
        ]),
        '#tree' => true,
      ];
      foreach ($attributes as $attribute) {
        $attributeValue = $this->configuration['attributes'][$attribute->getId()] ?? $this->sendEthicManager->getDefaultAttributeValue($attribute);
        if (in_array($attribute->getFormat(), ['SingleSelection', 'MultipleSelection'])) {
          $options = [];
          $attributeValue = is_string($attributeValue) ? array_filter(explode(',', $attributeValue)) : $attributeValue;
          foreach ($attribute->getOptions() as $option) {
            /** @var SendEthicAttributeOption $option */
            $options[$option->getId()] = $option->getName();
          }
          $form['sendethic_data']['attributes'][$attribute->getId()] = [
            '#type' => $this->moduleHandler->moduleExists('select2') ? 'select2' : 'select',
            '#title' => $attribute->getName(),
            '#options' => $options,
            '#empty_option' => $this->t('- None -'),
            '#description' => $attribute->getDescription(),
            '#default_value' => $attributeValue,
            '#multiple' => $attribute->getFormat() === 'MultipleSelection',
          ];
        } else {
          $form['sendethic_data']['attributes'][$attribute->getId()] = [
            '#type' => $attribute->getFormat() === 'Number' ? 'number' : 'textfield',
            '#title' => $attribute->getName(),
            '#description' => $attribute->getDescription(),
            '#default_value' => $attributeValue,
          ];
        }
      }
    } else {
      $form['sendethic_data']['attributes'] = [
        '#markup' => $this->t('You need to set an account first'),
      ];
    }

    $token_types = ['webform', 'webform_submission'];
    $form['sendethic_data']['token_tree_link'] = $this->buildTokenTreeElement($token_types);

    // Submission data.
    $form['submission_data'] = [
      '#type' => 'details',
      '#title' => $this->t('Submission data'),
    ];

    $form['submission_data']['excluded_data'] = [
      '#type' => 'webform_excluded_columns',
      '#title' => $this->t('Posted data'),
      '#title_display' => 'invisible',
      '#webform_id' => $webform->id(),
      '#required' => TRUE,
      '#default_value' => $this->configuration['excluded_data'],
    ];

    $this->elementTokenValidate($form);

    return $this->setSettingsParents($form);
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    parent::submitConfigurationForm($form, $form_state);
    $this->configuration['attributes'] = $form_state->getValue(['sendethic_data', 'attributes']);
    $this->applyFormStateToConfiguration($form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function postSave(WebformSubmissionInterface $webform_submission, $update = TRUE) {
    $state = $webform_submission->getWebform()->getSetting('results_disabled') ? WebformSubmissionInterface::STATE_COMPLETED : $webform_submission->getState();
    if ($state === WebformSubmissionInterface::STATE_COMPLETED) {
      // send ethic manager call
      // $data = $webform_submission->toArray(TRUE);
      $data = [
        'contact_mail' => $this->configuration['contact_mail'],
        'contact_firstname' => $this->configuration['contact_firstname'],
        'contact_lastname' => $this->configuration['contact_lastname'],
        'attributes' => $this->configuration['attributes'],
        'data' => [],
      ];

      // Default mail
      $data['contact_mail'] = $data['contact_mail'] === self::DEFAULT_EMAIL_VALUE ? '[site:name]' : $data['contact_mail'];

      // Remove unsupported properties from data.
      // These are typically added by other module's like metatag.
      $unsupported_properties = array_combine($this->unsupportedProperties, $this->unsupportedProperties);
      $data = array_diff_key($data, $unsupported_properties);

      // Flatten data and prioritize the element data over the
      // webform submission data.
      $element_data = $data['data'];
      unset($data['data']);
      $data = $element_data + $data;

      // Replace tokens.
      $data = $this->replaceTokens($data, $webform_submission);

      // Append custom data.
      if (!empty($this->configuration['custom_data'])) {
        $custom_data = Yaml::decode($this->configuration['custom_data']);
        // Replace tokens.
        $custom_data = $this->replaceTokens($custom_data, $webform_submission);
        $data = $custom_data + $data;
      }

      $contact = new SendEthicContact($data['contact_mail'], $data['contact_firstname'], $data['contact_lastname'], $data['attributes']);
      $account = $this->sendEthicManager->getDefaultAccount();

      // for debug/error
      $request_url = sprintf('%s/contact/attribute', SendEthicManager::BASE_URL);
      $request_method = 'POST';
      $request_type = 'json';
      $request_options = $contact->toRequestArray();

      // add
      try {
        $response = $this->sendEthicManager->addContact($contact, $account);
      }
      catch (RequestException $request_exception) {
        $response = $request_exception->getResponse();

        // Encode HTML entities to prevent broken markup from breaking the page.
        $message = $request_exception->getMessage();
        $message = nl2br(htmlentities($message));

        $this->handleError($state, $message, $request_url, 'POST', $request_type, [], $response);
        return;
      }

      if (!($response instanceof ResponseInterface)) {
        $this->handleError($state, 'Bad account', $request_url, 'POST', $request_type, [], $response);
        return;
      }

      // Display submission exception if response code is not 2xx.
      if ($this->responseHasError($response)) {
        $message = $this->t('Remote post request return @status_code status code.', ['@status_code' => $response->getStatusCode()]);
        $this->handleError($state, $message, $request_url, $request_method, $request_type, $request_options, $response);
        return;
      }
      else {
        $this->displayCustomResponseMessage($response, FALSE);
        // @TODO: log history POST /contact/history { "contactKey": "test@test.com", "logs": [ { "Title": "TEST", "LogDate": "YYYY-MM-DD HH:00:00+0200" } ] }
      }

      // If debugging is enabled, display the request and response.
      $this->debug($this->t('Remote post successful!'), $state, $request_url, $request_method, $request_type, $request_options, $response, 'warning');

      // Replace [webform:handler] tokens in submission data.
      // Data structured for [webform:handler:remote_post:completed:key] tokens.
      $submission_data = $webform_submission->getData();
      $submission_has_token = (strpos(print_r($submission_data, TRUE), '[webform:handler:' . $this->getHandlerId() . ':') !== FALSE) ? TRUE : FALSE;
      if ($submission_has_token) {
        $response_data = $request_options;
        $token_data = ['webform_handler' => [$this->getHandlerId() => [$state => $response_data]]];
        $submission_data = $this->replaceTokens($submission_data, $webform_submission, $token_data);
        $webform_submission->setData($submission_data);
        // Resave changes to the submission data without invoking any hooks
        // or handlers.
        $webform_submission->resave();
      }
    }
  }

  /**
   * Build A select other element for email address and names.
   *
   * @param string $name
   *   The element's key.
   * @param string $title
   *   The element's title.
   * @param string $label
   *   The element's label.
   * @param bool $required
   *   TRUE if the element is required.
   * @param array $element_options
   *   The element options.
   * @param array $options_options
   *   The options options.
   * @param array $role_options
   *   The (user) role options.
   * @param array $other_options
   *   The other options.
   *
   * @return array
   *   A select other element.
   */
  protected function buildElement($name, $title, $label, $required = FALSE, array $element_options = [], array $other_options = NULL) {
    [$element_name, $element_type] = (strpos($name, '_') !== FALSE) ? explode('_', $name) : [$name, 'text'];

    $default_option = '';

    $options = [];
    $options[WebformSelectOther::OTHER_OPTION] = $this->t('Custom @label…', ['@label' => $label]);
    if ($default_option) {
      $options[(string) $this->t('Default')] = [static::DEFAULT_EMAIL_VALUE => $default_option];
    }
    if (!empty($element_options)) {
      $options[(string) $this->t('Elements')] = $element_options;
    }
    if ($other_options) {
      $options[(string) $this->t('Other')] = $other_options;
    }

    $element = [];

    $element[$name] = [
      '#type' => 'webform_select_other',
      '#title' => $title,
      '#options' => $options,
      '#empty_option' => (!$required) ? $this->t('- None -') : NULL,
      '#other__title' => $title,
      '#other__title_display' => 'invisible',
      '#other__placeholder' => $this->t('Enter @label…', ['@label' => $label]),
      '#other__type' => ($element_type === 'mail') ? 'webform_email_multiple' : 'textfield',
      '#other__allow_tokens' => TRUE,
      '#required' => $required,
      '#default_value' => $this->configuration[$name],
    ];

    // Tweak elements.
    switch ($name) {
      case 'contact_mail':
        $element[$name]['#other__description'] = $this->t('Multiple email addresses may be separated by commas. Emails are only sent to cc and bcc addresses if a To email address is provided.')
          . ' '
          . $this->t("If multiple email addresses are entered the '@name' will be not included in the email.", ['@name' => $this->t('From name')]);
        break;
    }

    return $element;
  }

  /* ************************************************************************ */
  // Debug and exception handlers.
  /* ************************************************************************ */

  /**
   * Display debugging information.
   *
   * @param string $message
   *   Message to be displayed.
   * @param string $state
   *   The state of the webform submission.
   *   Either STATE_NEW, STATE_DRAFT_CREATED, STATE_DRAFT_UPDATED,
   *   STATE_COMPLETED, STATE_UPDATED, or STATE_CONVERTED
   *   depending on the last save operation performed.
   * @param string $request_url
   *   The remote URL the request is being posted to.
   * @param string $request_method
   *   The method of remote post.
   * @param string $request_type
   *   The type of remote post.
   * @param string $request_options
   *   The requests options including the submission data.
   * @param \Psr\Http\Message\ResponseInterface|null $response
   *   The response returned by the remote server.
   * @param string $type
   *   The type of message to be displayed to the end use.
   */
  protected function debug($message, $state, $request_url, $request_method, $request_type, $request_options, ResponseInterface $response = NULL, $type = 'warning') {
    if (empty($this->configuration['debug'])) {
      return;
    }

    $build = [
      '#type' => 'details',
      '#title' => $this->t('Debug: Remote post: @title [@state]', ['@title' => $this->label(), '@state' => $state]),
    ];

    // State.
    $build['state'] = [
      '#type' => 'item',
      '#title' => $this->t('Submission state/operation:'),
      '#markup' => $state,
      '#wrapper_attributes' => ['class' => ['container-inline'], 'style' => 'margin: 0'],
    ];

    // Request.
    $build['request'] = ['#markup' => '<hr />'];
    $build['request_url'] = [
      '#type' => 'item',
      '#title' => $this->t('Request URL'),
      '#markup' => $request_url,
      '#wrapper_attributes' => ['class' => ['container-inline'], 'style' => 'margin: 0'],
    ];
    $build['request_method'] = [
      '#type' => 'item',
      '#title' => $this->t('Request method'),
      '#markup' => $request_method,
      '#wrapper_attributes' => ['class' => ['container-inline'], 'style' => 'margin: 0'],
    ];
    $build['request_type'] = [
      '#type' => 'item',
      '#title' => $this->t('Request type'),
      '#markup' => $request_type,
      '#wrapper_attributes' => ['class' => ['container-inline'], 'style' => 'margin: 0'],
    ];
    $build['request_options'] = [
      '#type' => 'item',
      '#title' => $this->t('Request options'),
      '#wrapper_attributes' => ['style' => 'margin: 0'],
      'data' => [
        '#markup' => Html::escape(Yaml::encode($request_options)),
        '#prefix' => '<pre>',
        '#suffix' => '</pre>',
      ],
    ];

    // Response.
    $build['response'] = ['#markup' => '<hr />'];
    if ($response) {
      $build['response_code'] = [
        '#type' => 'item',
        '#title' => $this->t('Response status code:'),
        '#markup' => $response->getStatusCode(),
        '#wrapper_attributes' => ['class' => ['container-inline'], 'style' => 'margin: 0'],
      ];
      $build['response_header'] = [
        '#type' => 'item',
        '#title' => $this->t('Response header:'),
        '#wrapper_attributes' => ['style' => 'margin: 0'],
        'data' => [
          '#markup' => Html::escape(Yaml::encode($response->getHeaders())),
          '#prefix' => '<pre>',
          '#suffix' => '</pre>',
        ],
      ];
      $build['response_body'] = [
        '#type' => 'item',
        '#wrapper_attributes' => ['style' => 'margin: 0'],
        '#title' => $this->t('Response body:'),
        'data' => [
          '#markup' => Html::escape($response->getBody()),
          '#prefix' => '<pre>',
          '#suffix' => '</pre>',
        ],
      ];
    }
    else {
      $build['response_code'] = [
        '#markup' => $this->t('No response. Please see the recent log messages.'),
        '#prefix' => '<p>',
        '#suffix' => '</p>',
      ];
    }

    // Message.
    $build['message'] = ['#markup' => '<hr />'];
    $build['message_message'] = [
      '#type' => 'item',
      '#wrapper_attributes' => ['style' => 'margin: 0'],
      '#title' => $this->t('Message:'),
      '#markup' => $message,
    ];

    $this->messenger()->addMessage($this->renderer->renderPlain($build), $type);
  }

  /**
   * Handle error by logging and display debugging and/or exception message.
   *
   * @param string $state
   *   The state of the webform submission.
   *   Either STATE_NEW, STATE_DRAFT_CREATED, STATE_DRAFT_UPDATED,
   *   STATE_COMPLETED, STATE_UPDATED, or STATE_CONVERTED
   *   depending on the last save operation performed.
   * @param string $message
   *   Message to be displayed.
   * @param string $request_url
   *   The remote URL the request is being posted to.
   * @param string $request_method
   *   The method of remote post.
   * @param string $request_type
   *   The type of remote post.
   * @param string $request_options
   *   The requests options including the submission data.
   * @param \Psr\Http\Message\ResponseInterface|null $response
   *   The response returned by the remote server.
   */
  protected function handleError($state, $message, $request_url, $request_method, $request_type, $request_options, $response) {
    global $base_url, $base_path;

    // If debugging is enabled, display the error message on screen.
    $this->debug($message, $state, $request_url, $request_method, $request_type, $request_options, $response, 'error');

    // Log error message.
    $context = [
      '@form' => $this->getWebform()->label(),
      '@state' => $state,
      '@type' => $request_type,
      '@url' => $request_url,
      '@message' => $message,
      'webform_submission' => $this->getWebformSubmission(),
      'handler_id' => $this->getHandlerId(),
      'operation' => 'error',
      'link' => $this->getWebform()
        ->toLink($this->t('Edit'), 'handlers')
        ->toString(),
    ];
    $this->getLogger('webform_submission')
      ->error('@form webform remote @type post (@state) to @url failed. @message', $context);

    // Display custom or default exception message.
    if (!$this->displayCustomResponseMessage($response, TRUE)) {
      $this->messageManager->display(WebformMessageManagerInterface::SUBMISSION_EXCEPTION_MESSAGE, 'error');
    }

    // Redirect the current request to the error url.
    $error_url = $this->replaceTokens($this->configuration['error_url'], $this->getWebformSubmission());
    if ($error_url && PHP_SAPI !== 'cli') {
      // Convert error path to URL.
      if (strpos($error_url, '/') === 0) {
        $error_url = $base_url . preg_replace('#^' . $base_path . '#', '/', $error_url);
      }

      $request = $this->requestStack->getCurrentRequest();

      // Build Ajax redirect or trusted redirect response.
      $wrapper_format = $request->get(MainContentViewSubscriber::WRAPPER_FORMAT);
      $is_ajax_request = ($wrapper_format === 'drupal_ajax');
      if ($is_ajax_request) {
        $response = new AjaxResponse();
        $response->addCommand(new RedirectCommand($error_url));
        $response->setData($response->getCommands());
      }
      else {
        $response = new TrustedRedirectResponse($error_url);
      }
      // Save the session so things like messages get saved.
      $request->getSession()->save();
      $response->prepare($request);
      // Make sure to trigger kernel events.
      $this->kernel->terminate($request, $response);
      $response->send();
      // Only exit, an Ajax request to prevent headers from being overwritten.
      if ($is_ajax_request) {
        exit;
      }
    }
  }

  /**
   * Get custom response message.
   *
   * @param \Psr\Http\Message\ResponseInterface|null $response
   *   The response returned by the remote server.
   * @param bool $default
   *   Display the default message. Defaults to TRUE.
   *
   * @return string
   *   A custom response message.
   */
  protected function getCustomResponseMessage($response, $default = TRUE) {
    if (!empty($this->configuration['messages']) && $response instanceof ResponseInterface) {
      $status_code = $response->getStatusCode();
      foreach ($this->configuration['messages'] as $message_item) {
        if ((int) $message_item['code'] === (int) $status_code) {
          return $this->replaceTokens($message_item['message'], $this->getWebformSubmission());
        }
      }
    }
    return (!empty($this->configuration['message']) && $default)
      ? $this->replaceTokens($this->configuration['message'], $this->getWebformSubmission())
      : '';
  }

  /**
   * Display custom response message.
   *
   * @param \Psr\Http\Message\ResponseInterface|null $response
   *   The response returned by the remote server.
   * @param bool $default
   *   Display the default message. Defaults to TRUE.
   *
   * @return bool
   *   TRUE if custom response message is displayed.
   */
  protected function displayCustomResponseMessage($response, $default = TRUE) {
    $custom_response_message = $this->getCustomResponseMessage($response, $default);
    if (!$custom_response_message) {
      return FALSE;
    }

    $token_data = [
      'webform_handler' => [
        $this->getHandlerId() => [],//$this->getResponseData($response),
      ],
    ];
    $build_message = [
      '#markup' => $this->replaceTokens($custom_response_message, $this->getWebform(), $token_data),
    ];
    $message = \Drupal::service('renderer')->renderPlain($build_message);
    $type = ($this->responseHasError($response)) ? MessengerInterface::TYPE_ERROR : MessengerInterface::TYPE_STATUS;
    $this->messenger()->addMessage($message, $type);
    return TRUE;
  }

  /**
   * Determine if response has an error status code.
   *
   * @param \Psr\Http\Message\ResponseInterface|null $response
   *   The response returned by the remote server.
   *
   * @return bool
   *   TRUE if response status code reflects an unsuccessful value.
   */
  protected function responseHasError($response) {
    $status_code = $response->getStatusCode();
    return $status_code < 200 || $status_code >= 300;
  }

}
