<?php

namespace Drupal\sendethic\Service;

use Drupal\Component\Datetime\TimeInterface;
use Drupal\Component\Serialization\Json;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Language\LanguageManagerInterface;
use Drupal\Core\State\StateInterface;
use Drupal\sendethic\SendEthicAccount;
use Drupal\sendethic\SendEthicAttribute;
use Drupal\sendethic\SendEthicAttributeOption;
use Drupal\sendethic\SendEthicContact;
use Drupal\sendethic\SendEthicSegment;
use GuzzleHttp\ClientInterface;
use GuzzleHttp\Exception\RequestException;
use Psr\Http\Message\ResponseInterface;
use Psr\Log\LoggerInterface;

class SendEthicManager {

  /**
   * See: https://www.sendethic.com/campus/api-emailing-connecteurs/api/documentation-api-rest/
   */
  const BASE_URL = 'https://services.message-business.com/api/rest/v4';

  /**
   * The sendethic.settings config object.
   *
   * @var \Drupal\Core\Config\Config
   */
  protected $config;

  /**
   * The language manager object.
   *
   * @var \Drupal\Core\Language\LanguageManager
   */
  protected $languageManager;

  /**
   * The state object.
   *
   * @var \Drupal\Core\State\State
   */
  protected $state;

  /**
   * A logger instance.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected $logger;

  /**
   * The time service.
   *
   * @var \Drupal\Component\Datetime\TimeInterface
   */
  protected $time;

  /**
   * The HTTP client to fetch the feed data with.
   *
   * @var \GuzzleHttp\ClientInterface
   */
  protected $httpClient;

  /**
   * Constructs a XmlSitemapGenerator object.
   *
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The config factory object.
   * @param \Drupal\Core\State\StateInterface $state
   *   The state handler.
   * @param \Drupal\Core\Language\LanguageManagerInterface $language_manager
   *   Language Manager.
   * @param \Psr\Log\LoggerInterface $logger
   *   A logger instance.
   * @param \Drupal\Component\Datetime\TimeInterface $time
   *   The time service.
   * @param \GuzzleHttp\ClientInterface $http_client
   * The HTTP client to fetch the feed data with.
   */
  public function __construct(ConfigFactoryInterface $config_factory, StateInterface $state, LanguageManagerInterface $language_manager, LoggerInterface $logger, TimeInterface $time, ClientInterface $http_client) {
    $this->config = $config_factory->get('sendethic.settings');
    $this->state = $state;
    $this->languageManager = $language_manager;
    $this->logger = $logger;
    $this->time = $time;
    $this->httpClient = $http_client;
  }

  public function getDefaultAccount(): ?SendEthicAccount {
    if ($this->config->get('default.account_id') && $this->config->get('default.api_key')) {
      return new SendEthicAccount($this->config->get('default.account_id'), $this->config->get('default.api_key'));
    }
    return null;
  }

  /** @return array */
  protected function getRequestOptions(SendEthicAccount $account): array {
    return [
      'auth' => [ $account->getAccountId(), $account->getApiKey() ],
      'headers' => [
        'User-Agent' => 'MessageBusinessApi/1; PHP v. ' . phpversion(),
      ],
    ];
  }

  public function addContact(SendEthicContact $contact, ?SendEthicAccount $account = null): ?ResponseInterface {
    $account = $account ?: $this->getDefaultAccount();
    if ($account instanceof SendEthicAccount) {
      $request_options = $this->getRequestOptions($account);
      $request_options['json'] = [ 'contactKey' => 'email' ] + $contact->toRequestArray();
      return $this->httpClient->post(sprintf('%s/contact/attribute', self::BASE_URL), $request_options);
    }
    return null;
  }

  /** @return SendEthicSegment[] */
  public function getSegments(?SendEthicAccount $account = null): array {
    $account = $account ?: $this->getDefaultAccount();
    $segments = [];
    if ($account instanceof SendEthicAccount) {
      $request_options = $this->getRequestOptions($account);
      $response = $this->httpClient->get(sprintf('%s/segment', self::BASE_URL), $request_options);
      if ($response instanceof ResponseInterface && $response->getStatusCode() === 200) {
        foreach (Json::decode($response->getBody()->getContents()) as $segment_array) {
          $segment = new SendEthicSegment($segment_array['id'], $segment_array['name'], $segment_array['description'], $segment_array['details'], $segment_array['mode']);
          $segments[$segment->getId()] = $segment;
        }
      }
    }
    return $segments;
  }

  /** @return SendEthicAttribute[] */
  public function getAttributes(?SendEthicAccount $account = null): array {
    $account = $account ?: $this->getDefaultAccount();
    $attributes = [];
    if ($account instanceof SendEthicAccount) {
      $request_options = $this->getRequestOptions($account);
      $response = $this->httpClient->get(sprintf('%s/account/contactattribute', self::BASE_URL), $request_options);
      if ($response instanceof ResponseInterface && $response->getStatusCode() === 200) {
        foreach (Json::decode($response->getBody()->getContents()) as $attribute_array) {
          $options = null;
          if (in_array($attribute_array['format'], ['SingleSelection', 'MultipleSelection'])) {
            $options = [];
            $response = $this->httpClient->get(sprintf('%s/account/contactattribute/value/%d', self::BASE_URL, $attribute_array['id']), $request_options);
            if ($response instanceof ResponseInterface && $response->getStatusCode() === 200) {
              foreach (Json::decode($response->getBody()->getContents()) as $option_array) {
                $option = new SendEthicAttributeOption($option_array['id'], $option_array['text']);
                $options[$option->getId()] = $option;
              }
            }
          }
          $attribute = new SendEthicAttribute($attribute_array['id'], $attribute_array['name'], $attribute_array['description'], $attribute_array['format'], $options);
          $attributes[$attribute->getId()] = $attribute;
        }
      }
    }
    return $attributes;
  }

  public function getDefaultAttributeValue(SendEthicAttribute $attribute, ?SendEthicAccount $account = null) {
    $account = $account ?: $this->getDefaultAccount();
    if ($account instanceof SendEthicAccount) {
      return $this->config->get(sprintf('default.attributes.%s', $attribute->getId()));
    }
    return null;
  }

}
